/* gnubeep 1.4 - simple beep generation program
Copyright (C) 2003  Shaun Colley

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


/* Include files */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/io.h>
#include <signal.h>
/* #include "soundlib.h" */


/* Defines - ON and OFF for speaker() */
#define ON 1
#define OFF 0


/* this global is needed to determine whether the speaker is on 
 * speaker_on is set to 1 when the speaker is on, and 0 when the speaker
 * is off.  this way, we can keep track of whether the speaker is on or off -
 * this is important when signals are received, and the program needs to 
 * exit without leaving the speaker on (the tone would play forever)
 */
int speaker_on = 0;

/* function prototypes */
int tone(int frequency, int duration);
int speaker(int status);
int nosound();
int bell();
int usage();


/* tone generation function - 
 *
 * arguments: frequency
 *            duration
 *
 * frequency is the pitch of the tone,
 * duration is the length of the tone
 */
int tone(int frequency, int duration) {

	/* This calculation determines the number the PIT needs to be set to */
        int divisor = 1193180 / frequency;

	speaker_on = 1;  /* speaker is on */
	
	/* get permission to use the I/O ports 
	 * this requires root in linux, unless the SUID bit is set */
        if (ioperm(0x61, 0x20, 1)) {
		perror("ioperm()");
		return(1);
	}
	
	/* grab access to the rest of the I/O ports */
        ioperm(0x42, 0x02, 1);
        ioperm(0x43, 0x02, 1);

        /* Turn on the speaker */
	speaker(ON);

	/* Play the sound */
        outb(0xB6, 0x43);
	
	/* The value which channel 2 of the PIT is set to is a word, so
	 * the value must be sent in two lots: */
        outb(divisor & 0xff, 0x42);  /* Send the first byte. */
        outb(divisor >> 8, 0x42);  /* Shift by 8 bits to second part of word. */


	/* Sleep for specified duration */
        usleep(duration);

	/* Stop the sound */
        nosound();

        return(0);
}


/* sound control function -
 *
 * arguments: none
 */
int nosound() {

	/* turn off the speaker */
        speaker(OFF);

	/* set speaker_on back to 0, because the speaker is now off */
	speaker_on = 0;

	/* give back the I/O ports */
        if (ioperm(0x61, 0x20, 0)) {
		perror("ioperm()");
		return 1;
	}
	
	/* give back other I/O ports */
        ioperm(0x42, 0x02, 0);
        ioperm(0x43, 0x02, 0);
	
	return(0);
}


/* speaker control function - 
 *
 * arguments: status - 0 or 1 (on/off)
 * the arguments specifies what status the speaker should be set to
 */
int speaker(int status) {

        int portval;

	/* Get current port val */
        portval = inb(0x61);

	/* theck how the program wants the speaker */
        if(status == ON) {
                portval = portval | 3;  /* this will set the speaker on */
        }

        else {
                portval = portval & 0xfc;  /* this should turn speaker off */
        }

	
	/* Send the value to speaker */
        outb(portval, 0x61);
        return(0);
}

/* print the 'bell' character - generate a standard beep */
int bell() {
	printf("\a");
	return(0);
}

/* a simple usage summary display function */
int usage() {
printf("gnubeep 1.4\n");
printf("Written by Shaun Colley\n\n");
printf("Usage: gnubeep [-flbh]\n\n");
printf("\t-f        <frequency> Frequency of tone\n");
printf("\t-l        <length> Length of tone\n");
printf("\t-b        Generate standard beep ('bell' character)\n");
printf("\t-h        Displays this help output\n\n\n");
printf("See manpage for more info\n");

/* exit with error code 1 */
return(1);
}


/* signal handler.
 * This is very important because if the program was terminated via a signal,
 * the speaker would NOT be turned off, and the tone would continue to play
 * until the system is rebooted.  to prevent this problem, I have implemented
 * a simple signal handler, which turns the speaker off it necessary, and 
 * exits.  before doing anything, signals are ignored to prevent signal races.
 *
 * arguments: none
 */
void sighdlr() {

	/* ignore signals - hopefully prevent signal races */
	signal(SIGINT, SIG_IGN);
	signal(SIGQUIT, SIG_IGN);
        signal(SIGABRT, SIG_IGN);
	signal(SIGSTOP, SIG_IGN);
	signal(SIGTSTP, SIG_IGN);
	signal(SIGTERM, SIG_IGN);
	signal(SIGHUP, SIG_IGN);

	/* is the speaker on? 
	 * when tone() is called, speaker_on is set to 1.  when nosound()
	 * is called, speaker_on is reset to 0 again.  this system allows
	 * us to keep track of the speaker status, which is important in the
	 * case of needed to exit, because we need to turn the speaker off,
	 * it necessary - otherwise, the tone would sound until the system is
	 * rebooted! 
	 */
	if(speaker_on == 1)
		nosound();

	exit(1);
}


int main(int argc, char *argv[]) {

        int x;
	
	/* these variables are used to house the tone frequency and length.
	 * they are initialised at 0 so it is easy to tell if a user has
	 * specified them or not.  if they did not specify them as arguments,
	 * they will still be equal to 0 after the getopt loop. */
	long int dur = 0;
	int freq = 0; 
	
	/* install signal handlers */
	signal(SIGINT, sighdlr);
	signal(SIGQUIT, sighdlr);
        signal(SIGABRT, sighdlr);
	signal(SIGSTOP, sighdlr);
	signal(SIGTSTP, sighdlr);
	signal(SIGTERM, sighdlr);
	signal(SIGHUP, sighdlr);
									       	
	/* check for arguments, set variables appropriately */
	while ((x = getopt (argc, argv, "f:l:bh")) != EOF) {
		
		/* parse arguments */
		switch(x) {
			case 'h' : usage();  /* display help output */
				   return(0);
				   break;

			case 'f' : freq = atoi(optarg);  /* specify the tone
							    frequency */
				   break;

			case 'l':  dur = atol(optarg);  /* specify the tone
							   duration */
				   break;

			case 'b' : bell();  /* play the standard bell sound - 
					       equivalent to 'printf("\a")' */
				   return(0);
				   break;

			default : usage();  /* no valid options */
				  exit(-1);
				  break;
		} /* end switch case */
	} /* end while */
	
	
	/* has the user specified a frequency and duration? */
	if(freq == 0 || dur == 0) {
		usage();
		exit(-1);
	}
	
	/* generate tone */
	tone(freq, dur);
	
	/* exit - success */
	return(0);
}
